import torch

from torch import Tensor
from torch.nn import Module


class BatchRenorm(Module):
    def __init__(
        self,
        num_features: int,
        epsilon: float = 1e-3,
        momentum: float = 0.01,
        affine: bool = True,
    ) -> None:

        super().__init__()

        self.register_buffer('running_mean', torch.zeros(num_features, dtype=torch.float32))
        self.register_buffer('running_std', torch.ones(num_features, dtype=torch.float32))
        self.register_buffer('num_batches_tracked', torch.tensor(0, dtype=torch.int64))

        self.weight = torch.nn.Parameter(torch.ones(num_features, dtype=torch.float32))
        self.bias = torch.nn.Parameter(torch.zeros(num_features, dtype=torch.float32))

        self.affine = affine
        self.epsilon = epsilon
        self.step = 0
        self.momentum = momentum

    def _check_input_dim(self, x):
        raise NotImplementedError()

    @property
    def rmax(self):
        return (2 / 35000 * self.num_batches_tracked + 25 / 35).clamp_(1.0, 3.0)

    @property
    def dmax(self):
        return (5 / 20000 * self.num_batches_tracked - 25 / 20).clamp_(0.0, 5.0)

    def forward(self, x):
        self._check_input_dim(x)

        if x.dim() > 2:
            x = x.transpose(1, -1)

        if self.training:
            dims = [i for i in range(x.dim() - 1)]
            batch_mean = x.mean(dims)
            batch_std = x.std(dims, unbiased=False) + self.epsilon
            r = (
                batch_std.detach() / self.running_std.view_as(batch_std)
            ).clamp_(1 / self.rmax, self.rmax)
            d = (
                (batch_mean.detach() - self.running_mean.view_as(batch_mean)) / self.running_std.view_as(batch_std)
            ).clamp_(-self.dmax, self.dmax)
            x = (x - batch_mean) / batch_std * r + d
            self.running_mean += self.momentum * (batch_mean.detach() - self.running_mean)
            self.running_std += self.momentum * (batch_std.detach() - self.running_std)
            self.num_batches_tracked += 1
        else:
            x = (x - self.running_mean) / self.running_std

        if self.affine:
            x = self.weight * x + self.bias

        if x.dim() > 2:
            x = x.transpose(1, -1)

        return x


class BatchRenorm1d(BatchRenorm):
    def _check_input_dim(self, x: Tensor) -> None:
        if x.dim() not in [2, 3]:
            raise ValueError(f'expected 2D or 3D input (got {x.dim()}D input)')


class BatchRenorm2d(BatchRenorm):
    def _check_input_dim(self, x: Tensor) -> None:
        if x.dim() != 4:
            raise ValueError(f'expected 4D input (got {x.dim()}D input)')


class BatchRenorm3d(BatchRenorm):
    def _check_input_dim(self, x: Tensor) -> None:
        if x.dim() != 5:
            raise ValueError(f'expected 5D input (got {x.dim()}D input)')



import functools
from typing import Tuple, Sequence, List, Callable, Union, Optional

import numpy as np
import torch
import torchvision.transforms.functional as functional
from PIL.Image import Image
from torch import Tensor
from torch.nn import Module
from torchvision.transforms import RandomHorizontalFlip, ColorJitter, RandomResizedCrop, RandomRotation, GaussianBlur
from torchvision.transforms.functional import InterpolationMode

import sfu_torch_lib.tree as tree_lib


class ResizeTree(Module):
    def __init__(self, sizes: Sequence[Tuple[int, int]], interpolations: Sequence[InterpolationMode]) -> None:
        super().__init__()

        self.sizes = sizes
        self.interpolations = interpolations

    def forward(self, tree):
        return tree_lib.map_tree(
            lambda image, size, interpolation: functional.resize(image, size, interpolation),
            tree, self.sizes, self.interpolations,
        )


class RandomResizedCropTree(Module):
    def __init__(
            self,
            sizes: Sequence[Tuple[int, int]],
            interpolations: Sequence[InterpolationMode],
            scale: Tuple[float, float] = (0.08, 1.0),
            ratio: Tuple[float, float] = (3. / 4., 4. / 3.),
    ) -> None:

        super().__init__()

        self.sizes = sizes
        self.interpolations = interpolations
        self.scale = scale
        self.ratio = ratio

    def forward(self, tree):
        i, j, h, w = RandomResizedCrop.get_params(tree[0], self.scale, self.ratio)

        tree = tree_lib.map_tree(
            lambda image, size, interpolation: functional.resized_crop(image, i, j, h, w, size, interpolation),
            tree, self.sizes, self.interpolations,
        )

        return tree


class RandomHorizontalFlipTree(RandomHorizontalFlip):
    def forward(self, tree):
        flip = torch.rand(1) < self.p

        tree = tree_lib.map_tree(lambda image: functional.hflip(image) if flip else image, tree)

        return tree


class ColorJitterTree(ColorJitter):
    @staticmethod
    def transform(function_indices, brightness_factor, contrast_factor, saturation_factor, hue_factor, image):
        for function_index in function_indices:
            if function_index == 0 and brightness_factor is not None:
                image = functional.adjust_brightness(image, brightness_factor)
            elif function_index == 1 and contrast_factor is not None:
                image = functional.adjust_contrast(image, contrast_factor)
            elif function_index == 2 and saturation_factor is not None:
                image = functional.adjust_saturation(image, saturation_factor)
            elif function_index == 3 and hue_factor is not None:
                image = functional.adjust_hue(image, hue_factor)

        return image

    def forward(self, tree):
        transform = functools.partial(
            self.transform,
            *self.get_params(self.brightness, self.contrast, self.saturation, self.hue),
        )

        new_images = tree_lib.map_tree(transform, tree)

        return new_images


class ToTensorTree(Module):
    def __init__(self, *dtypes) -> None:
        super().__init__()

        self.dtypes = dtypes

    def forward(self, tree):
        return tree_lib.map_tree(lambda image, dtype: torch.as_tensor(np.array(image, dtype)), tree, self.dtypes)


class ConvertImageTree(Module):
    def __init__(self, modes: Sequence[Optional[str]]) -> None:
        super().__init__()

        self.modes = modes

    @staticmethod
    def transform(image: Image, mode: Optional[str]) -> Image:
        return image.convert(mode) if mode is not None else image

    def forward(self, tree):
        return tree_lib.map_tree(self.transform, tree, self.modes)


class EncodeImageTree(Module):
    def __init__(self, means: Optional[Tensor] = None, scales: Optional[Tensor] = None) -> None:
        super().__init__()

        self.means = means
        self.scales = scales

    def transform(self, tensor: Tensor) -> Tensor:
        tensor = torch.permute(tensor, (2, 0, 1))
        tensor -= self.means[:, None, None] if self.means is not None else 128
        tensor /= self.scales[:, None, None] if self.scales is not None else 128

        return tensor

    def forward(self, tree):
        return tree_lib.map_tree(self.transform, tree)


class RandomRotationTree(Module):
    def __init__(
            self,
            degrees: Union[float, Tuple[float, float]],
            interpolations: Sequence[InterpolationMode],
            expand: bool = False,
            center: Optional[Tuple[float, float]] = None,
            fill: Union[float, List[float]] = 0,
    ) -> None:

        super().__init__()

        self.degrees = [-degrees, degrees] if isinstance(degrees, float) else degrees
        self.interpolations = interpolations
        self.expand = expand
        self.center = center
        self.fill = fill

    def forward(self, tree):
        angle = RandomRotation.get_params(self.degrees)

        new_tree = tree_lib.map_tree(
            lambda image, interpolation: functional.rotate(
                img=image,
                angle=angle,
                interpolation=interpolation,
                expand=self.expand,
                center=self.center,
                fill=self.fill,
            ),
            tree, self.interpolations
        )

        return new_tree


class GaussianBlurTree(GaussianBlur):
    def forward(self, tree):
        sigma = self.get_params(self.sigma[0], self.sigma[1])

        new_tree = tree_lib.map_tree(
            lambda image, : functional.gaussian_blur(image, self.kernel_size, [sigma, sigma]),
            tree
        )

        return new_tree


class SelectTree(Module):
    def __init__(self, paths) -> None:
        super().__init__()

        self.paths = paths

    def forward(self, tree):
        return tree_lib.select_tree(tree, self.paths)


class ComposeTree(Module):
    def __init__(self, steps: List[Tuple[List[int], Callable]]) -> None:
        super().__init__()

        self.steps = steps

    def forward(self, tree):
        for paths, function in self.steps:
            tree = tree_lib.map_tree_subset(function, tree, paths)

        return tree



import logging
import os
from typing import Type, Dict, Any, Optional

import mlflow
import torch
from pytorch_lightning import LightningModule

import sfu_torch_lib.io as io


logger = logging.getLogger(__name__)


def get_checkpoint_path(run_id: str, filename: str = 'last') -> Optional[str]:
    run = mlflow.get_run(run_id)

    checkpoint_path = os.path.join(run.info.artifact_uri, f'{filename}.ckpt')

    if not io.exists(checkpoint_path):
        return None

    return checkpoint_path


def get_localized_checkpoint_path(run_id: str, filename: str = 'last') -> Optional[str]:
    checkpoint_path = get_checkpoint_path(run_id, filename)
    checkpoint_path = io.localize_file(checkpoint_path) if checkpoint_path else None

    return checkpoint_path


def get_resumable_checkpoint_path(
        run_id: Optional[str],
        restart: bool = False,
        filename: str = 'last',
) -> Optional[str]:

    if restart or not run_id:
        return None

    checkpoint_path = get_localized_checkpoint_path(run_id, filename)

    if not checkpoint_path:
        run = mlflow.get_run(run_id)

        if len(run.data.metrics) > 0:
            logging.warning('Could not find checkpoint. The model will not be initialized.')

    return checkpoint_path


def get_checkpoint(run_id: str, filename: str = 'last') -> Optional[Dict[str, Any]]:
    checkpoint_path = get_checkpoint_path(run_id, filename)

    if checkpoint_path is None:
        return None

    with io.open(checkpoint_path) as checkpoint_file:
        checkpoint = torch.load(checkpoint_file)

    return checkpoint


def load_model(run_id: str, module_class: Type[LightningModule], filename: str = 'last', **kwargs) -> LightningModule:
    checkpoint_path = get_checkpoint_path(run_id, filename)

    assert checkpoint_path

    with io.open(checkpoint_path) as checkpoint_file:
        model = module_class.load_from_checkpoint(checkpoint_file, **kwargs)

    assert isinstance(model, LightningModule)

    return model


def load_checkpoint_state(checkpoint_path: str, model: LightningModule) -> None:
    device = torch.device('cuda') if torch.cuda.device_count() else torch.device('cpu')

    with io.open(checkpoint_path) as checkpoint_file:
        checkpoint = torch.load(checkpoint_file, device)

    model.load_state_dict(checkpoint['state_dict'])


def load_state(run_id: str, model: LightningModule, filename: str = 'last') -> None:
    checkpoint_path = get_checkpoint_path(run_id, filename)

    assert checkpoint_path

    load_checkpoint_state(checkpoint_path, model)



import inspect
import sys
from argparse import Namespace
from typing import Callable, Optional, List, Mapping, Any, Dict

import sfu_torch_lib.utils as utils


def get_positional_arguments(argv: Optional[List[str]] = None) -> List[str]:
    argv = argv if argv else sys.argv[1:]

    arguments = []

    for argument in argv:
        if argument.startswith('-'):
            break

        arguments.append(argument)

    return arguments


def get_keyword_arguments(argv: Optional[List[str]] = None) -> Dict[str, str]:
    argv = argv if argv else sys.argv[1:]

    num_positional_arguments = len(get_positional_arguments(argv))

    passed_arguments = {
        key.lstrip('-').replace('_', '-'): value
        for key, value
        in utils.get_pairs(argv[num_positional_arguments:])
        if key.startswith('-')
    }

    return passed_arguments


def get_script_parameters(function: Callable, ignore_keyword_arguments: bool = True) -> Dict[str, Any]:
    """
    Returns the arguments of a function with its values specified by the command line or its default values.
    Underscores in the name of the arguments are transformed to dashes.
    Can optionally filter out keyword arguments obtained through the command line.

    :param function: The function to inspect.
    :param ignore_keyword_arguments: Whether to filter out keyword command line arguments.
    :return: A map from argument names to default values.
    """
    positional_arguments, keyword_arguments = get_positional_arguments(), get_keyword_arguments()
    signature = inspect.signature(function)

    arguments = {}

    for index, (name, parameter) in enumerate(signature.parameters.items()):
        transformed_name = name.replace('_', '-')

        if index < len(positional_arguments):
            arguments[transformed_name] = positional_arguments[index]

        if not ignore_keyword_arguments and transformed_name in keyword_arguments:
            arguments[transformed_name] = keyword_arguments[transformed_name]

        elif transformed_name not in keyword_arguments and parameter.default != parameter.empty:
            arguments[transformed_name] = parameter.default

    return arguments


def flatten_dictionary(parameters: Mapping[Any, Any], delimiter: str = '/') -> Dict[str, Any]:
    def _dictionary_generator(input_dictionary, prefixes=None):
        prefixes = prefixes[:] if prefixes else []

        if isinstance(input_dictionary, Mapping):
            for key, value in input_dictionary.items():
                key = str(key)

                if isinstance(value, (Mapping, Namespace)):
                    value = vars(value) if isinstance(value, Namespace) else value
                    yield from _dictionary_generator(value, prefixes + [key])

                else:
                    yield prefixes + [key, value if value is not None else str(None)]

        else:
            yield prefixes + [input_dictionary if input_dictionary is None else str(input_dictionary)]

    return {delimiter.join(keys): val for *keys, val in _dictionary_generator(parameters)}



import functools
from typing import Tuple, Sequence, Iterator, Optional, NamedTuple, Union, List

import torch
import torch.autograd as autograd
from torch import Tensor
from torch.optim import Optimizer


class LossArguments(NamedTuple):
    target_losses: Union[Tensor, Sequence[Tensor]]
    adaptive_loss: Optional[Tensor] = None
    scale_factor: float = 1.


def scale_tensor(tensor: Tensor, scale: Union[float, Tensor] = 1.) -> Tensor:
    return tensor * scale if scale != 1 else tensor


def get_parameters(optimizer: Optimizer) -> Iterator[Tensor]:
    yield from (
        parameter
        for parameter_group in optimizer.param_groups
        for _, parameters in parameter_group.items()
        for parameter in (parameters if isinstance(parameters, list) else [parameters])
        if torch.is_tensor(parameter) and parameter.requires_grad
    )


def manual_step(optimizer: Optimizer, loss: Tensor, allow_unused: bool = False) -> None:
    optimizer.zero_grad()

    parameters = list(get_parameters(optimizer))
    gradients = autograd.grad(loss, parameters, retain_graph=True, allow_unused=allow_unused)

    for parameter, gradient in zip(parameters, gradients):
        parameter.grad = gradient

    optimizer.step()


def calculate_norms(
        gradients_first: Sequence[Optional[Tensor]],
        gradients_second: Sequence[Optional[Tensor]],
) -> Tuple[Tensor, Tensor]:

    norm_first, norm_second = functools.reduce(
        lambda first, second: (first[0] + second[0], first[1] + second[1]),
        (
            (torch.sum(gradient_first ** 2), torch.sum(gradient_second ** 2))
            for gradient_first, gradient_second in zip(gradients_first, gradients_second)
            if gradient_first is not None and gradient_second is not None
        ),
    )

    norm_first, norm_second = torch.sqrt(norm_first), torch.sqrt(norm_second)

    return norm_first, norm_second


def combine_gradients(
        gradients_first: Sequence[Optional[Tensor]],
        gradients_second: Sequence[Optional[Tensor]],
        bidirectional: bool = True,
        scale_factor_first: Union[float, Tensor] = 1.,
        scale_factor_second: Union[float, Tensor] = 1.,
) -> Iterator[Optional[Tensor]]:

    norm_first, norm_second = calculate_norms(gradients_first, gradients_second)

    scale_factor_first *= torch.minimum(norm_first, norm_second) / norm_first
    scale_factor_first = scale_factor_first if bidirectional else 1.

    scale_factor_second *= torch.minimum(norm_first, norm_second) / norm_second

    for gradient_first, gradient_second in zip(gradients_first, gradients_second):
        if gradient_first is not None and gradient_second is not None:
            yield scale_tensor(gradient_first, scale_factor_first) + scale_tensor(gradient_second, scale_factor_second)

        elif gradient_first is not None:
            yield scale_tensor(gradient_first, scale_factor_first)

        elif gradient_second is not None:
            yield scale_tensor(gradient_second, scale_factor_second)

        else:
            yield None


def combine_gradients_losses(losses: Sequence[Tensor], parameters: Sequence[Tensor]) -> Iterator[Optional[Tensor]]:
    yield from functools.reduce(
        lambda first, second: combine_gradients(list(first), list(second)),  # type: ignore
        (
            autograd.grad(loss, parameters, retain_graph=True, allow_unused=True)
            for loss in losses
        ),
    )


def collect_gradients(parameters: Sequence[Tensor], losses: Sequence[LossArguments]) -> Iterator[List[Tensor]]:
    collected_gradients = []

    for loss in losses:
        target_losses = [loss.target_losses] if isinstance(loss.target_losses, Tensor) else loss.target_losses
        gradients = combine_gradients_losses(target_losses, parameters)

        if loss.adaptive_loss is None:
            gradients = (
                scale_tensor(gradient, loss.scale_factor) if gradient is not None else gradient
                for gradient in gradients
            )

        else:
            gradients = combine_gradients(
                gradients_first=list(gradients),
                gradients_second=autograd.grad(loss.adaptive_loss, parameters, retain_graph=True, allow_unused=True),
                bidirectional=False,
                scale_factor_second=loss.scale_factor,
            )

        collected_gradients.append(gradients)

    yield from (
        [gradient for gradient in gradients if gradient is not None]
        for gradients in zip(*collected_gradients)
    )


def step_adaptive_clipping(optimizer: Optimizer, losses: Sequence[LossArguments]) -> None:
    optimizer.zero_grad()

    parameters = list(get_parameters(optimizer))
    gradients = collect_gradients(parameters, losses)

    for parameter, gradient in zip(parameters, gradients):
        if len(gradient) > 0:
            parameter.grad = torch.add(*gradient)

    optimizer.step()



import functools
import logging
import os
import sys
import traceback
from logging import StreamHandler
from typing import Optional, Dict, Any, TextIO, Mapping, Callable

import mlflow
import pytorch_lightning.loggers.logger as logger_lib
from mlflow.entities import RunTag
from mlflow.tracking import MlflowClient
from mlflow.utils.mlflow_tags import MLFLOW_RUN_NAME
from pytorch_lightning import Trainer, LightningModule
from pytorch_lightning.callbacks import ModelCheckpoint
from pytorch_lightning.loggers import MLFlowLogger as MLFlowLoggerBase

import sfu_torch_lib.io as io
import sfu_torch_lib.parameters as parameters_lib
import sfu_torch_lib.utils as utils


class MLFlowLogger(MLFlowLoggerBase):
    def __init__(
        self,
        run_name: Optional[str] = None,
        tracking_uri: Optional[str] = os.getenv('MLFLOW_TRACKING_URI'),
        tags: Optional[Dict[str, Any]] = None,
        save_dir: Optional[str] = './mlruns',
        prefix: str = '',
        artifact_location: Optional[str] = None,
    ) -> None:

        super().__init__(
            run_name=run_name,
            tracking_uri=tracking_uri,
            tags=tags,
            save_dir=save_dir,
            prefix=prefix,
            artifact_location=artifact_location,
        )

        self.logger = logging.getLogger(__name__)

    @property  # type: ignore[misc]
    @logger_lib.rank_zero_experiment
    def experiment(self) -> MlflowClient:
        if self._run_id and self._experiment_id:
            return self._mlflow_client

        self.tags = self.tags or {}

        if self._run_name is not None:
            if MLFLOW_RUN_NAME in self.tags:
                self.logger.warning(
                    f'The tag {MLFLOW_RUN_NAME} is found in tags. '
                    f'The value will be overridden by {self._run_name}.'
                )

            self.tags[MLFLOW_RUN_NAME] = self._run_name

        run = mlflow.active_run() or mlflow.start_run()

        self._mlflow_client.log_batch(
            run_id=run.info.run_id,
            tags=[
                RunTag(key, str(value))
                for key, value in self.tags.items()
            ],
        )

        self._run_id = run.info.run_id
        self._experiment_id = run.info.experiment_id
        self._experiment_name = mlflow.get_experiment(run.info.experiment_id).name

        return self._mlflow_client


class MLFlowModelCheckpoint(ModelCheckpoint):
    def __init__(self, patience: int = 0, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.patience = patience
        self.mlflow_epoch = 0
        self.mlflow_model_path = ''

    @staticmethod
    def log_artifact(local_file: str, filename: str = 'last.ckpt') -> None:
        bucket, key = io.get_bucket_and_key(mlflow.get_artifact_uri())
        destination_path = os.path.join(key, filename)

        io.upload_s3(bucket, destination_path, local_file)

    def has_new_model(self) -> bool:
        return bool(self.best_model_path) and self.best_model_path != self.mlflow_model_path

    def update_model(self, trainer: Trainer) -> None:
        if self.has_new_model() and io.exists(self.best_model_path):
            self.log_artifact(self.best_model_path)
            self.mlflow_epoch = trainer.current_epoch
            self.mlflow_model_path = self.best_model_path

    def on_train_epoch_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        super().on_train_epoch_end(trainer, pl_module)

        if trainer.current_epoch - self.mlflow_epoch >= self.patience:
            self.update_model(trainer)

    def on_train_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        super().on_train_end(trainer, pl_module)
        self.update_model(trainer)

    def on_exception(self, trainer: Trainer, pl_module: LightningModule, exception: BaseException) -> None:
        super().on_exception(trainer, pl_module, exception)
        self.update_model(trainer)


class StreamPersister:
    def __init__(self, stream: TextIO) -> None:
        self.stream = stream
        self.path = io.generate_path(f'{stream.name.strip("<>")}.log', prefix=None)
        self.file = open(self.path, mode='w')
        self.encoding = 'utf-8'

    def write(self, message: str) -> None:
        self.stream.write(message)
        self.file.write(message)

    def flush(self) -> None:
        self.stream.flush()
        self.file.flush()


def add_params(function: Callable) -> Callable:
    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        parameters = parameters_lib.get_script_parameters(function, ignore_keyword_arguments=False)
        mlflow.log_params(parameters)

        return function(*args, **kwargs)

    return wrapper


def log(function: Callable, override: bool = utils.to_bool_or_false(os.getenv('DEVELOPMENT_MODE'))) -> Callable:
    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        if override:
            return function(*args, **kwargs)

        sys.stdout = StreamPersister(sys.stdout)
        sys.stderr = StreamPersister(sys.stderr)

        handler = StreamHandler(sys.stderr.file)

        for name in logging.root.manager.loggerDict:
            logger = logging.getLogger(name)

            if not logger.propagate:
                logger.addHandler(handler)

        try:
            return function(*args, **kwargs)

        except Exception as exception:
            sys.stderr.write(f'{exception}\n{traceback.format_exc()}')
            raise exception

        finally:
            sys.stdout.flush()
            sys.stderr.flush()
            mlflow.log_artifact(sys.stdout.path)
            mlflow.log_artifact(sys.stderr.path)

    return wrapper


def install(function: Callable, override: bool = utils.to_bool_or_false(os.getenv('DEVELOPMENT_MODE'))) -> Callable:
    return log(add_params(function), override)


def get_metrics(run_id: Optional[str] = None) -> Mapping[str, float]:
    run_id = run_id if run_id else mlflow.active_run().info.run_id
    metrics = mlflow.get_run(run_id).data.metrics if run_id else {}

    return metrics



import functools
import math
import re
from typing import (
    List, Tuple, TypeVar, Iterable, Iterator, Type, Union, Sequence, Mapping, Callable, Generic, Dict, Optional,
)

from torch.nn import ELU, Identity, ReLU, LeakyReLU, Sigmoid, Softmax, Tanh, Module
from torch.utils.data import Dataset


T = TypeVar('T')


class AttributeDict(Generic[T]):
    def __init__(self, dictionary: Dict[str, T]) -> None:
        self.dictionary = dictionary

        for key, value in self.items():
            setattr(self, key.lower().replace(' ', '_'), value)

    def __getattr__(self, name):
        return getattr(self, name)

    def __getitem__(self, key):
        return self.dictionary[key]

    def __repr__(self):
        return repr(self.dictionary)

    def __len__(self):
        return len(self.dictionary)

    def __contains__(self, item):
        return item in self.dictionary

    def __iter__(self):
        return iter(self.dictionary)

    def keys(self):
        return self.dictionary.keys()

    def values(self):
        return self.dictionary.values()

    def items(self):
        return self.dictionary.items()


def get_pairs(items: Iterable[T]) -> Iterator[Tuple[T, T]]:
    first = None

    for index, item in enumerate(items):
        if index % 2 == 0:
            first = item
        else:
            yield first, item  # type: ignore


def create_selector(indices: Union[int, Sequence[int]]) -> Callable[[Sequence[T]], Union[T, Tuple[T, ...]]]:
    def select(items: Sequence[T]) -> Union[T, Tuple[T, ...]]:
        if isinstance(indices, int):
            return items[indices]

        return tuple(items[index] for index in indices)

    return select


def to_list(*args: Union[Sequence[T], T]) -> List[T]:
    items: List[T] = []

    for argument in args:
        if isinstance(argument, Sequence):
            items += argument
        else:
            items.append(argument)

    return items


def to_bool(value: Optional[str]) -> Optional[bool]:
    if value is None:
        return None

    if value == '0' or value.lower() == 'false':
        return False

    return bool(value)


def to_bool_or_false(value: Optional[str]) -> bool:
    return to_bool(value) or False


def get_activation(name: str) -> Type[Module]:
    return {
        'elu': ELU,
        'identity': Identity,
        'relu': ReLU,
        'leaky_relu': LeakyReLU,
        'sigmoid': Sigmoid,
        'softmax': Softmax,
        'tanh': Tanh,
    }[name]


def parse_map_sequence(string: str, sequence_type: Type[T]) -> Mapping[str, Sequence[T]]:
    substrings = re.split(r'(?<=]),', string)

    output = {}

    for substring in substrings:
        key, value = substring.split(':')

        output[key] = list(map(sequence_type, value.strip('[]').split(',')))

    return output


def get_num_steps(dataset: Dataset, batch_size: int, frequency: int = 10) -> int:
    return max(1, math.floor(len(dataset) / batch_size / frequency))  # type: ignore


parse_map_ints = functools.partial(parse_map_sequence, sequence_type=int)
parse_map_floats = functools.partial(parse_map_sequence, sequence_type=float)
parse_map_strings = functools.partial(parse_map_sequence, sequence_type=str)



import glob
import io
import os
import pickle
import shutil
import tempfile
from contextlib import contextmanager
from io import BytesIO
from typing import Optional, List, Tuple, Union, Generator, IO, Any, Literal
from urllib.parse import urlparse

import boto3
import numpy as np
from s3fs import S3FileSystem

import sfu_torch_lib.utils as utils


ModeType = Literal['r', 'r+', 'w+', 'c']
MMapModeType = Literal['readonly', 'r', 'copyonwrite', 'c', 'readwrite', 'r+', 'write', 'w+']


def get_client(endpoint_url: Optional[str] = os.getenv('S3_ENDPOINT_URL'), verify: Optional[bool] = None):
    verify = utils.to_bool(os.getenv('S3_VERIFY')) if verify is None else verify

    client = boto3.client('s3', endpoint_url=endpoint_url, verify=verify)

    return client


def get_s3fs(endpoint_url: Optional[str] = os.getenv('S3_ENDPOINT_URL'), verify: Optional[bool] = None) -> S3FileSystem:
    verify = utils.to_bool(os.getenv('S3_VERIFY')) if verify is None else verify

    filesystem = S3FileSystem(client_kwargs={'endpoint_url': endpoint_url, 'verify': verify})

    return filesystem


def get_bucket_and_key(path: str) -> Tuple[str, str]:
    url = urlparse(path)
    bucket, key = url.netloc, url.path.strip('/')

    return bucket, key


def download_s3(bucket: str, key: str, destination: str) -> None:
    directory, _ = os.path.split(destination)

    os.makedirs(directory, exist_ok=True)

    get_client().download_file(Bucket=bucket, Key=key, Filename=destination)


def upload_s3(bucket: str, key: str, source: Union[str, BytesIO]) -> None:
    client = get_client()

    if isinstance(source, str):
        client.upload_file(Filename=source, Bucket=bucket, Key=key)
    else:
        client.upload_fileobj(Fileobj=source, Bucket=bucket, Key=key)


def _get_s3_keys(bucket: str, prefix: str = '') -> Generator[str, None, None]:
    object_list = get_client().list_objects_v2(Bucket=bucket, Prefix=prefix)

    if 'Contents' not in object_list:
        return

    for file in object_list['Contents']:
        yield file['Key']

    while object_list['IsTruncated']:
        object_list = get_client().list_objects_v2(
            Bucket=bucket,
            Prefix=prefix,
            ContinuationToken=object_list['NextContinuationToken'],
        )

        if 'Contents' not in object_list:
            return

        for file in object_list['Contents']:
            yield file['Key']


def _get_local_files(path: str) -> List[str]:
    paths = [
        path
        for path in glob.glob(os.path.join(path, '**'), recursive=True)
        if os.path.isfile(path)
    ]

    return paths


def get_files(path: str) -> List[str]:
    if not path.startswith('s3'):
        return _get_local_files(path)

    bucket, prefix = get_bucket_and_key(path)

    paths = [f's3://{os.path.join(bucket, key)}' for key in _get_s3_keys(bucket, prefix)]

    return paths


def exists(path: str) -> bool:
    if not path.startswith('s3'):
        return os.path.exists(path)

    bucket, key = get_bucket_and_key(path)
    has_keys = sum(1 for _ in _get_s3_keys(bucket, key)) > 0

    return has_keys


def generate_path(suffix: Optional[str] = None, prefix: Optional[str] = os.getenv('DATA_PATH')) -> str:
    prefix = prefix if prefix is not None else tempfile.gettempdir()

    if not suffix:
        return prefix

    path = os.path.join(prefix, suffix)

    return path


def localize_directory(
        path: str,
        directory: Optional[str] = os.getenv('DATA_EPHEMERAL_PATH'),
        prefix: Optional[str] = None,
        overwrite: bool = True,
) -> str:

    basename = os.path.basename(path)

    target = generate_path(basename) if directory is None else generate_path(basename, directory)
    target_prefixed = os.path.join(target, prefix) if prefix else target

    if not overwrite and exists(target_prefixed):
        return target

    elif path.startswith('s3'):
        cache = generate_path(basename)
        cache_prefixed = os.path.join(cache, prefix) if prefix else cache

        if overwrite or not exists(cache):
            bucket, key_base = get_bucket_and_key(path)
            key_base_prefixed = os.path.join(key_base, prefix) if prefix else key_base
            key_base_prefixed = key_base_prefixed if key_base_prefixed.endswith('/') else key_base_prefixed + '/'

            for key in _get_s3_keys(bucket, key_base_prefixed):
                download_s3(bucket, key, generate_path(os.path.relpath(key, key_base), cache))

        if cache != target:
            shutil.copytree(cache_prefixed, target_prefixed)

        return target

    elif directory is not None:
        path_prefixed = os.path.join(path, prefix) if prefix else path

        if path != target:
            shutil.copytree(path_prefixed, target_prefixed)

        return target

    else:
        return path


def localize_file(
        path: str,
        directory: Optional[str] = os.getenv('DATA_EPHEMERAL_PATH'),
        overwrite: bool = True,
) -> str:

    basename = os.path.basename(path)
    target = generate_path(basename) if directory is None else generate_path(basename, directory)

    if not overwrite and exists(target):
        return target

    elif path.startswith('s3'):
        cache = generate_path(basename)

        if overwrite or not exists(cache):
            bucket, key = get_bucket_and_key(path)
            download_s3(bucket, key, cache)

        if cache != target:
            shutil.copy(cache, target)

        return target

    elif directory is not None:
        if path != target:
            shutil.copy(path, target)

        return target

    else:
        return path


def localize(
        path: str,
        directory: Optional[str] = os.getenv('DATA_EPHEMERAL_PATH'),
        overwrite: bool = True,
        override: bool = utils.to_bool_or_false(os.getenv('DEVELOPMENT_MODE')),
) -> str:

    if override:
        return path.split('.')[0]
    elif '.' in os.path.basename(path):
        return localize_file(path, directory, overwrite)
    else:
        return localize_directory(path, directory, overwrite=overwrite)


def load_npy(path: str, mmap_mode: Optional[ModeType] = None) -> np.ndarray:
    return np.load(localize_file(path), mmap_mode=mmap_mode)


def load_bin(
        path: str,
        dtype=np.uint8,
        mode: MMapModeType = 'r+',
        shape: Optional[Union[int, Tuple[int, ...]]] = None,
) -> np.ndarray:

    return np.memmap(localize_file(path), dtype, mode, shape=shape)


def open(path: str, mode: str = 'rb', encoding: Optional[str] = None, newline: Optional[str] = None) -> IO:
    open_function = get_s3fs().open if path.startswith('s3') else io.open

    file_object = open_function(path, mode=mode, encoding=encoding, newline=newline)

    return file_object


def read(path: str, encoding: Optional[str] = None) -> Union[bytes, str]:
    if not path.startswith('s3'):
        if encoding:
            return io.open(path, mode='r', encoding=encoding).read()
        else:
            return io.open(path, mode='rb').read()

    bucket, key = get_bucket_and_key(path)

    content = get_client().get_object(Bucket=bucket, Key=key)['Body'].read()

    if encoding:
        return content.decode(encoding)

    return content


@contextmanager
def with_local_file(path: str) -> Generator[str, None, None]:
    local_path = generate_path(os.path.basename(path)) if path.startswith('s3') else path

    yield local_path

    if path.startswith('s3'):
        bucket, key = get_bucket_and_key(path)
        upload_s3(bucket, key, local_path)


@contextmanager
def with_local_directory(path: str) -> Generator[str, None, None]:
    local_path = generate_path(os.path.basename(path)) if path.startswith('s3') else path

    yield local_path

    if path.startswith('s3'):
        sync_remote(local_path, path)


def save_npy(path: str, data: np.ndarray) -> None:
    if not path.startswith('s3'):
        np.save(path, data)

    else:
        with open(path, mode='wb') as writer:
            np.save(writer, data)


def save_pickle(path: str, data: Any) -> None:
    if not path.startswith('s3'):
        with io.open(path, mode='wb') as file:
            pickle.dump(data, file)

    else:
        with open(path, mode='wb') as writer:
            pickle.dump(data, writer)


def sync_remote(local_path: str, remote_path: str) -> None:
    paths = (path for path in glob.glob(os.path.join(local_path, '**'), recursive=True) if os.path.isfile(path))

    for path in paths:
        path_in_remote_path = generate_path(os.path.relpath(path, local_path), remote_path)

        bucket, key = get_bucket_and_key(path_in_remote_path)
        upload_s3(bucket, key, path)



from typing import Sequence


def copy_tree(tree):
    tree_copy = []
    queue = [(tree, tree_copy)]

    while queue:
        source_node, destination_node = queue.pop(0)

        for source_child in source_node:
            if isinstance(source_child, Sequence):
                new_node = []
                destination_node.append(new_node)
                queue.append((source_child, new_node))
            else:
                destination_node.append(source_child)

    return tree_copy


def zip_tree(*trees):
    tree_new = []
    queue = [(trees, tree_new)]

    while queue:
        source_nodes, destination_node = queue.pop(0)

        for source_children in zip(*source_nodes):
            if isinstance(source_children[0], Sequence):
                new_node = []
                destination_node.append(new_node)
                queue.append((source_children, new_node))
            else:
                destination_node.append(source_children)

    return tree_new


def flatten_tree(data, paths):
    data_flat = []

    for path in paths:
        node = data

        path = [path] if isinstance(path, int) else path

        for index in path:
            node = node[index]

        data_flat.append(node)

    return data_flat


def unflatten_tree(tree, data_flat, paths):
    new_tree = copy_tree(tree)

    for datum, path in zip(data_flat, paths):
        node = new_tree

        path = [path] if isinstance(path, int) else path

        for index in path[:-1]:
            node = node[index]

        node[path[-1]] = datum

    return new_tree


def map_tree(function, *trees):
    new_tree = zip_tree(*trees)

    queue = [new_tree]

    while queue:
        node = queue.pop(0)

        for index, child in enumerate(node):
            if isinstance(child, list):
                queue.append(child)
            else:
                arguments = child if isinstance(child, tuple) else tuple(child)
                node[index] = function(*arguments)

    return new_tree


def select_tree(tree, mask):
    tree_copy = []

    forward_stack = [(tree, mask, tree_copy)]
    backward_stack = []

    while forward_stack:
        source_node, mask_node, destination_node = forward_stack.pop()

        for source_child, mask_child in zip(source_node, mask_node):
            if isinstance(source_child, Sequence):
                new_node = []
                forward_stack.append((source_child, mask_child, new_node))
                backward_stack.append((new_node, destination_node))
            elif mask_child:
                backward_stack.append((source_child, destination_node))

    while backward_stack:
        child, parent = backward_stack.pop()

        if child:
            parent.insert(0, child)

    return tree_copy


def map_tree_subset(function, data, paths):
    leaves = flatten_tree(data, paths)
    results = function(leaves)
    output = unflatten_tree(data, results, paths)

    return output
